
#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif

//#include <linux/config.h>
#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/mm.h>
#include <linux/dma-mapping.h>
#include <asm/uaccess.h>
#include "prefixes/prefixes.c"
// I like the next two lines
#include "../devel/hijack.h"
#include "../devel/hijack.c"

struct hijack *phijack_dma_map_sg, *phijack_dma_unmap_sg;

static struct page *thepage;
static long prev_dma;
static long phys_addr;
static int n_dma_map_single, n_dma_map_sg;
int my_dma_map_sg(struct device *dev, struct scatterlist *sg, int nents,
		  enum dma_data_direction direction)
{
	int ret;
	int i;
	int (*orig) (struct device * dev, struct scatterlist * sg,
		     int nents, enum dma_data_direction direction) =
	    get_hijack_tramp(phijack_dma_map_sg);
//	n_dma_map_sg++;
	ret = orig(dev, sg, nents, direction);
	for (i = 0; i < nents; i++)
		if (sg[i].page == thepage) {
			printk("got page, objname=%s\n", dev->kobj.k_name);
			prev_dma = sg[i].dma_address;
			if (phys_addr)
				sg[i].dma_address = phys_addr;
			else
				printk("assign phys_addr first!\n");
		}
	return ret;
}
void
my_dma_unmap_sg(struct device *hwdev, struct scatterlist *sg, int nents,
		enum dma_data_direction direction)
{
	void (*orig) (struct device * dev, struct scatterlist * sg,
		      int nents, enum dma_data_direction direction) =
	    get_hijack_tramp(phijack_dma_unmap_sg);
	int i;
	for (i = 0; i < nents; i++)
		if (sg[i].page == thepage) {
			printk("got page in unmap, objname=%s\n",
			       hwdev->kobj.k_name);
			sg[i].dma_address = prev_dma;
			thepage = 0;
		}

	orig(hwdev, sg, nents, direction);
}

static int
ncall_read_proc(char *buf, char **start, off_t offset,
		int count, int *eof, void *data)
{
	int len = sprintf(buf, "single=0x%x sg=0x%x",
			  n_dma_map_single, n_dma_map_sg);
	*eof = 1;
	return len;
}
static int ncall_write_proc(struct file *file,
			    const char *buffer,
			    unsigned long count, void *data)
{
	long addr;
	if (copy_from_user(&addr, buffer, sizeof(addr)))
		return -EFAULT;
	if (get_user_pages(current, current->mm, (unsigned long) addr, 1,
			   1, 1, &thepage, NULL) != 1) {
		printk("get_user_pages error\n");
		return -1;
	}
	printk("userbuf=0x%lx, page=0x%p\n", addr, thepage);
	return sizeof(addr);
}
static int physaddr_write_proc(struct file *file,
			       const char *buffer,
			       unsigned long count, void *data)
{
	if (copy_from_user(&phys_addr, buffer, sizeof(phys_addr)))
		return -EFAULT;
	else
		return sizeof(phys_addr);
}
static long stext=-1;
module_param(stext, long, 0);
MODULE_PARM_DESC(stext, "please pass the \"stext\" value from /proc/kallsyms");
void * scratch;
int direct_init_module(void)
{
	int ret;
	struct proc_dir_entry *pbuf, *ncall;

	if (stext==-1) {
		printk("please pass the \"stext\" argument (value from /proc/kallsyms)");
		return -EINVAL;
	}
	ncall=create_proc_entry("userbuf", 0666, NULL);
	ncall->read_proc = ncall_read_proc;
	ncall->write_proc = ncall_write_proc;
	pbuf = create_proc_entry("physaddr", 0666, NULL);
	pbuf->write_proc = physaddr_write_proc;
        scratch=__vmalloc(8192, GFP_KERNEL | __GFP_HIGHMEM,
                __pgprot(__PAGE_KERNEL & ~_PAGE_NX)
                );
	phijack_dma_map_sg=scratch;
	phijack_dma_unmap_sg=(long)scratch+100;
	ret=init_hijack(phijack_dma_map_sg, dma_map_sg, my_dma_map_sg,dma_map_sg_prefix);
	ret=init_hijack(phijack_dma_unmap_sg, dma_unmap_sg, my_dma_unmap_sg,dma_unmap_sg_prefix);
		
	do_hijack_aux(phijack_dma_map_sg, stext);
	do_hijack_aux(phijack_dma_unmap_sg, stext+8);

	return 0;
}

void direct_cleanup_module(void)
{
	undo_hijack(phijack_dma_map_sg);
	undo_hijack(phijack_dma_unmap_sg);
	remove_proc_entry("userbuf", NULL);
	remove_proc_entry("physaddr", NULL);
	vfree(scratch);
}


module_init(direct_init_module);
module_exit(direct_cleanup_module);
